import asyncio
from datetime import datetime, timezone
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import eth_account
from bidict import bidict
from web3 import Web3

try:
    from web3.middleware import geth_poa_middleware
except ImportError:
    from web3.middleware import ExtraDataToPOAMiddleware as geth_poa_middleware

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.tegro import tegro_constants as CONSTANTS, tegro_utils, tegro_web_utils as web_utils
from hummingbot.connector.exchange.tegro.tegro_api_order_book_data_source import TegroAPIOrderBookDataSource
from hummingbot.connector.exchange.tegro.tegro_api_user_stream_data_source import TegroUserStreamDataSource
from hummingbot.connector.exchange.tegro.tegro_auth import TegroAuth
from hummingbot.connector.exchange.tegro.tegro_messages import encode_typed_data
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, to_0x_hex
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import DeductedFromReturnsTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter

s_logger = None
s_decimal_0 = Decimal(0)
s_float_NaN = float("nan")
MAX_UINT256 = 2**256 - 1


class TegroExchange(ExchangePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0

    web_utils = web_utils

    def __init__(self,
                 client_config_map: "ClientConfigAdapter",
                 tegro_api_key: str,
                 tegro_api_secret: str,
                 chain_name: str = CONSTANTS.DEFAULT_CHAIN,
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN
                 ):
        self.api_key = tegro_api_key
        self._chain = chain_name
        self.secret_key = tegro_api_secret
        self._api_factory = WebAssistantsFactory
        self._domain = domain
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_trades_poll_tegro_timestamp = 1.0
        super().__init__(client_config_map)
        self._allowance_polling_task: Optional[asyncio.Task] = None
        self.real_time_balance_update = False

    @property
    def authenticator(self):
        return TegroAuth(
            api_key=self.api_key,
            api_secret=self.secret_key
        )

    @property
    def name(self) -> str:
        return self._domain

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def node_rpc(self):
        return f"tegro_{self._chain}_testnet" if self._domain.endswith("_testnet") else self._chain

    @property
    def chain_id(self):
        return self._chain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def chain(self):
        chain = 8453
        if self._domain.endswith("_testnet"):
            chain = CONSTANTS.TESTNET_CHAIN_IDS[self.chain_id]
        elif self._domain == "tegro":
            chain_id = CONSTANTS.DEFAULT_CHAIN
            # In this case tegro is default to base mainnet
            chain = CONSTANTS.MAINNET_CHAIN_IDS[chain_id]
        return chain

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.PING_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    async def _get_all_pairs_prices(self) -> Dict[str, Any]:
        results = {}
        pairs_prices = await self._api_get(
            path_url=CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL.format(self.chain),
            params={"page": 1, "sort_order": "desc", "sort_by": "volume", "page_size": 20, "verified": "true"},
            limit_id=CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL
        )
        for pair_price_data in pairs_prices:
            results[pair_price_data["symbol"]] = {
                "best_bid": pair_price_data["ticker"]["price"],
                "best_ask": pair_price_data["ticker"]["price"],
            }
        return results

    async def get_all_pairs_prices(self) -> Dict[str, Any]:
        res = []
        pairs_prices = await self._api_get(
            path_url=CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL.format(self.chain),
            params={"page": 1, "sort_order": "desc", "sort_by": "volume", "page_size": 20, "verified": "true"},
            limit_id=CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL
        )
        for pair_price_data in pairs_prices:
            result = {}
            result["symbol"] = pair_price_data["symbol"]
            result["price"] = pair_price_data["ticker"]["price"]
            res.append(result)
        return res

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_description = str(request_exception)
        is_time_synchronizer_related = ("-1021" in error_description
                                        and "Timestamp for this request" in error_description)
        return is_time_synchronizer_related

    def _is_request_result_an_error_related_to_time_synchronizer(self, request_result: Dict[str, Any]) -> bool:
        # The exchange returns a response failure and not a valid response
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return str(CONSTANTS.ORDER_NOT_EXIST_ERROR_CODE) in str(
            status_update_exception
        ) and CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return str(CONSTANTS.UNKNOWN_ORDER_ERROR_CODE) in str(
            cancelation_exception
        ) and CONSTANTS.UNKNOWN_ORDER_MESSAGE in str(cancelation_exception)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return TegroAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return TegroUserStreamDataSource(
            auth=self._auth,
            domain=self.domain,
            throttler=self._throttler,
            api_factory=self._web_assistants_factory,
        )

    def _get_fee(
        self,
        base_currency: str,
        quote_currency: str,
        order_type: OrderType,
        order_side: TradeType,
        amount: Decimal,
        price: Decimal = s_decimal_NaN,
        is_maker: Optional[bool] = None,
    ) -> TradeFeeBase:
        is_maker = True if is_maker is None else is_maker
        return DeductedFromReturnsTradeFee(percent=self.estimate_fee_pct(is_maker))

    async def start_network(self):
        await super().start_network()
        self._allowance_polling_task = safe_ensure_future(self.approve_allowance())

    async def stop_network(self):
        await super().stop_network()
        if self._allowance_polling_task is not None:
            self._allowance_polling_task.cancel()
            self._allowance_polling_task = None

    async def get_chain_list(self):
        account_info = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.CHAIN_LIST,
            limit_id=CONSTANTS.CHAIN_LIST,
            is_auth_required=False)

        return account_info

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:
        transaction_data = await self._generate_typed_data(amount, order_type, price, trade_type, trading_pair)
        s = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        symbol: str = s.replace('-', '_')
        signature = self.sign_inner(transaction_data)
        api_params = {
            "chain_id": self.chain,
            "base_asset": transaction_data["limit_order"]["base_asset"],
            "quote_asset": transaction_data["limit_order"]["quote_asset"],
            "side": transaction_data["limit_order"]["side"],
            "volume_precision": transaction_data["limit_order"]["volume_precision"],
            "price_precision": transaction_data["limit_order"]["price_precision"],
            "order_hash": transaction_data["limit_order"]["order_hash"],
            "raw_order_data": transaction_data["limit_order"]["raw_order_data"],
            "signature": signature,
            "signed_order_type": "tegro",
            "market_id": transaction_data["limit_order"]["market_id"],
            "market_symbol": symbol,
        }
        try:
            data = await self._api_request(
                path_url = CONSTANTS.ORDER_PATH_URL,
                method = RESTMethod.POST,
                data = api_params,
                is_auth_required = False,
                limit_id = CONSTANTS.ORDER_PATH_URL)
        except IOError as e:
            error_description = str(e)
            insufficient_allowance = ("insufficient allowance" in error_description)
            is_server_overloaded = ("status is 503" in error_description
                                    and "Unknown error, please check your request or try again later." in error_description)
            if insufficient_allowance:
                await self.approve_allowance(token=symbol)
            if is_server_overloaded:
                o_id = "Unknown"
                transact_time = int(datetime.now(timezone.utc).timestamp() * 1e3)
            else:
                raise
        else:
            o_id = f"{data['order_id']}"
            transact_time = data["timestamp"] * 1e-3
        return o_id, transact_time

    async def _generate_typed_data(self, amount, order_type, price, trade_type, trading_pair) -> Dict[str, Any]:
        side_str = CONSTANTS.SIDE_BUY if trade_type is TradeType.BUY else CONSTANTS.SIDE_SELL
        params = {
            "chain_id": self.chain,
            "market_symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
            "side": side_str,
            "wallet_address": self.api_key
        }
        data = await self.initialize_verified_market()
        # quote_precision = int(data["quote_precision"])
        base_precision = int(data["base_precision"])
        price_precision = 18
        params["amount"] = f"{amount:.{base_precision}g}"
        if order_type is OrderType.LIMIT or order_type is OrderType.LIMIT_MAKER:
            price_str = price
            params["price"] = f"{price_str:.{price_precision}g}"
        try:
            return await self._api_request(
                path_url = CONSTANTS.GENERATE_SIGN_URL,
                method = RESTMethod.POST,
                data = params,
                is_auth_required = False,
                limit_id = CONSTANTS.GENERATE_SIGN_URL)
        except IOError as e:
            raise IOError(f"Error submitting order {e}")

    async def _generate_cancel_order_typed_data(self, order_id: str, ids: list) -> Dict[str, Any]:
        try:
            params = {
                "order_ids": ids,
                "user_address": self.api_key.lower()
            }
            data = await self._api_request(
                path_url=CONSTANTS.GENERATE_ORDER_URL,
                method=RESTMethod.POST,
                data=params,
                is_auth_required=False,
                limit_id=CONSTANTS.GENERATE_ORDER_URL,
            )
            return self.sign_inner(data)
        except IOError as e:
            error_description = str(e)
            is_not_active = ("Orders not found" in error_description)
            if is_not_active:
                self.logger().debug(f"The order {order_id} does not exist on tegro."
                                    f"No cancelation needed.")
                return "Order not found"
            else:
                raise

    def sign_inner(self, data):
        message = "Order" if "Order" in data["sign_data"]["types"] else "CancelOrder"
        domain_data = data["sign_data"]["domain"]
        message_data = data["sign_data"]["message"]
        message_types = {message: data["sign_data"]["types"][message]}
        # encode and sign
        structured_data = encode_typed_data(domain_data, message_types, message_data)
        return to_0x_hex(eth_account.Account.from_key(self.secret_key).sign_message(structured_data).signature)

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        ids = []
        ex_oid = (await tracked_order.get_exchange_order_id()).split("+")[0]
        ids.append(ex_oid)
        signature = await self._generate_cancel_order_typed_data(order_id, ids)
        if signature is not None and signature != "Order not found":
            params = {
                "user_address": self.api_key,
                "order_ids": ids,
                "Signature": signature,
            }
            cancel_result = await self._api_request(
                path_url=CONSTANTS.CANCEL_ORDER_URL,
                method=RESTMethod.POST,
                data=params,
                is_auth_required=False,
                limit_id=CONSTANTS.CANCEL_ORDER_URL)
            result = cancel_result["cancelled_order_ids"][0]
            return True if result == ids[0] else False
        elif signature == "Order not found":
            await self._order_tracker.process_order_not_found(order_id)

    async def _format_trading_rules(self, exchange_info: List[Dict[str, Any]]) -> List[TradingRule]:
        """
        Example:
            {
                "id": "80002_0xfd655398df1c2e40c383b022fba15751e8e2ab49_0x7551122e441edbf3fffcbcf2f7fcc636b636482b",
                "symbol": "AYB_USDT",
                "chainId": 80002,
                "state": "verified",
                "base_contract_address": "0xfd655398df1c2e40c383b022fba15751e8e2ab49",
                "base_symbol": "AYB",
                "base_decimal": 18,
                "base_precision": 0,
                "quote_contract_address": "0x7551122e441edbf3fffcbcf2f7fcc636b636482b",
                "quote_symbol": "USDT",
                "quote_decimal": 6,
                "quote_precision": 10,
                "ticker": {
                    "base_volume": 0,
                    "quote_volume": 0,
                    "price": 0,
                    "price_change_24h": 0,
                    "price_high_24h": 0,
                    "price_low_24h": 0,
                    "ask_low": 0,
                    "bid_high": 0
                }
            }
        """
        retval = []
        for rule in filter(tegro_utils.is_exchange_information_valid, exchange_info):
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule.get("symbol"))
                min_order_size = Decimal(f'1e-{rule["base_precision"]}')
                min_price_inc = Decimal(f"1e-{rule['quote_precision']}")
                step_size = Decimal(f'1e-{rule["base_precision"]}')
                retval.append(
                    TradingRule(trading_pair,
                                min_order_size = Decimal(min_order_size),
                                min_price_increment = Decimal(min_price_inc),
                                min_base_amount_increment = Decimal(step_size)))
            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {rule}. Skipping.")
        return retval

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _user_stream_event_listener(self):
        """
        Listens to messages from _user_stream_tracker.user_stream queue.
        Traders, Orders, and Balance updates from the WS.
        """
        user_channels = CONSTANTS.USER_METHODS
        async for event_message in self._iter_user_event_queue():
            try:
                channel: str = event_message.get("action", None)
                results: Dict[str, Any] = event_message.get("data", {})
                if "code" not in event_message and channel not in user_channels.values():
                    self.logger().error(
                        f"Unexpected message in user stream: {event_message}.", exc_info = True)
                    continue
                elif channel == CONSTANTS.USER_METHODS["ORDER_SUBMITTED"]:
                    await self._process_order_message(results)
                elif channel == CONSTANTS.USER_METHODS["ORDER_TRADE_PROCESSED"]:
                    await self._process_order_message(results, fetch_trades = True)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    def _create_order_update_with_order_status_data(self, order_status: Dict[str, Any], order: InFlightOrder):
        new_states = self.get_state(order_status)
        confirmed_state = CONSTANTS.ORDER_STATE[new_states]
        order_update = OrderUpdate(
            trading_pair=order.trading_pair,
            update_timestamp=order_status["timestamp"] * 1e-3,
            new_state=confirmed_state,
            client_order_id=order.client_order_id,
            exchange_order_id=f"{str(order_status['order_id'])}",
        )
        return order_update

    async def _process_order_message(self, raw_msg: Dict[str, Any], fetch_trades = False):
        client_order_id = f"{raw_msg['order_id']}"
        tracked_order = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(client_order_id)
        if not tracked_order:
            self.logger().debug(f"Ignoring order message with id {client_order_id}: not in in_flight_orders.")
            return
        if fetch_trades:
            # process trade fill
            await self._all_trade_updates_for_order(order=tracked_order)
        order_update = self._create_order_update_with_order_status_data(order_status=raw_msg, order=tracked_order)
        self._order_tracker.process_order_update(order_update=order_update)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = (await order.get_exchange_order_id()).split("+")[0]
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            all_fills_response = await self._api_get(
                path_url=CONSTANTS.TRADES_FOR_ORDER_PATH_URL.format(exchange_order_id),
                is_auth_required=False,
                limit_id=CONSTANTS.TRADES_FOR_ORDER_PATH_URL)

            if len(all_fills_response) > 0:
                for trade in all_fills_response:
                    timestamp = trade["timestamp"]
                    symbol = trade["symbol"].split('_')[1]
                    fees = "0"
                    if order.trade_type == TradeType.BUY:
                        fees = trade["maker_fee"] if trade["is_buyer_maker"] else trade["taker_fee"]
                    if order.trade_type == TradeType.SELL:
                        fees = trade["taker_fee"] if trade["is_buyer_maker"] else trade["maker_fee"]
                    fee = TradeFeeBase.new_spot_fee(
                        fee_schema = self.trade_fee_schema(),
                        trade_type = order.trade_type,
                        percent_token = symbol,
                        flat_fees = [TokenAmount(amount=Decimal(fees), token=symbol)]
                    )
                    trade_update = TradeUpdate(
                        trade_id=trade["id"],
                        client_order_id=order.client_order_id,
                        exchange_order_id=order.exchange_order_id,
                        trading_pair=trading_pair,
                        fee=fee,
                        fill_base_amount=Decimal(trade["amount"]),
                        fill_quote_amount=Decimal(trade["amount"]) * Decimal(trade["price"]),
                        fill_price=Decimal(trade["price"]),
                        fill_timestamp=timestamp * 1e-3)
                    self._order_tracker.process_trade_update(trade_update)
                    trade_updates.append(trade_update)

        return trade_updates

    def get_state(self, updated_order_data):
        new_states = ""
        data = {}
        if isinstance(updated_order_data, list):
            state = updated_order_data[0]["status"]
            data = updated_order_data[0]
        else:
            state = updated_order_data["status"]
            data = updated_order_data
        if state == "closed" and Decimal(data["quantity_pending"]) == Decimal("0"):
            new_states = "completed"
        elif state == "open" and Decimal(data["quantity_filled"]) < Decimal("0"):
            new_states = "open"
        elif state == "open" and Decimal(data["quantity_filled"]) > Decimal("0"):
            new_states = "partial"
        elif state == "closed" and Decimal(data["quantity_pending"]) > Decimal("0"):
            new_states = "pending"
        elif state == "cancelled" and data["cancel"]["code"] == 611:
            new_states = "cancelled"
        elif state == "cancelled" and data["cancel"]["code"] != 611:
            new_states = "failed"
        else:
            new_states = data["status"]
        return new_states

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        o_id = (await tracked_order.get_exchange_order_id()).split("+")[0]
        updated_order_data = await self._api_get(
            path_url=CONSTANTS.TEGRO_USER_ORDER_PATH_URL.format(self.api_key),
            params = {
                "chain_id": self.chain,
                "order_id": o_id
            },
            limit_id=CONSTANTS.TEGRO_USER_ORDER_PATH_URL,
            is_auth_required=False)
        new_states = self.get_state(updated_order_data)
        confirmed_state = CONSTANTS.ORDER_STATE[new_states]
        order_update = OrderUpdate(
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=tracked_order.exchange_order_id,
            trading_pair=tracked_order.trading_pair,
            update_timestamp=updated_order_data[0]["timestamp"] * 1e-3,
            new_state=confirmed_state)
        return order_update

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        balances = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.ACCOUNTS_PATH_URL.format(self.chain, self.api_key),
            limit_id=CONSTANTS.ACCOUNTS_PATH_URL,
            is_auth_required=False)

        for balance_entry in balances:
            asset_name = balance_entry["symbol"]
            bal = float(str(balance_entry["balance"]))
            balance = Decimal(bal)
            self._account_available_balances[asset_name] = balance
            self._account_balances[asset_name] = balance
            remote_asset_names.add(asset_name)
        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset in asset_names_to_remove:
            del self._account_available_balances[asset]
            del self._account_balances[asset]

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: list[Dict[str, Any]]):
        mapping = bidict()
        for symbol_data in exchange_info:
            if tegro_utils.is_exchange_information_valid(exchange_info=symbol_data):
                try:
                    base, quote = symbol_data['symbol'].split('_')
                    mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(base=base, quote=quote)
                except Exception as exception:
                    self.logger().error(f"There was an error parsing a trading pair information ({exception})")
        self._set_trading_pair_symbol_map(mapping)

    async def approve_allowance(self, token=None, fail_silently: bool = True):
        """
        Approves the allowance for a specific token on a decentralized exchange.

        This function retrieves the trading pairs, determines the associated
        symbols, and approves the maximum allowance for each token in the
        trading pairs on the specified exchange contract.

        Returns:
        dict or None: The transaction receipt if the transaction is successful, otherwise None.
        """
        exchange_con_addr = ""
        token_list = []
        data = {}

        # Fetching trading pairs and determining associated symbols
        for trading_pair in self.trading_pairs:
            symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
            base, quote = symbol.split("_")
            token_list.append(base)
            token_list.append(quote)

        # If a specific token is provided, use only that token
        if token:
            base, quote = token.split("_")
            token_list = [base, quote]

        # Setting up Web3
        w3 = Web3(Web3.HTTPProvider(CONSTANTS.Node_URLS[self.node_rpc]))
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        approve_abi = CONSTANTS.ABI["approve"]

        # Fetching token and chain information
        tokens = await self.tokens_info()
        chain_data = await self.get_chain_list()
        exchange_con_addr = [
            Web3.to_checksum_address(chain["exchange_contract"])
            for chain in chain_data if int(chain["id"]) == self.chain][0]
        receipts = []
        # Organizing token data
        for t in tokens:
            data[t["symbol"]] = {"address": t["address"]}
        # Loop through each token and approve allowance

        for token in token_list:
            con_addr = Web3.to_checksum_address(data[token]["address"])
            addr = Web3.to_checksum_address(self.api_key)
            contract = w3.eth.contract(con_addr, abi=approve_abi)
            # Get nonce
            nonce = w3.eth.get_transaction_count(addr)
            # Prepare transaction parameters
            tx_params = {"from": addr, "nonce": nonce, "gasPrice": w3.eth.gas_price}
            try:
                # Estimate gas for the approval transaction
                gas_estimate = contract.functions.approve(exchange_con_addr, MAX_UINT256).estimate_gas({
                    "from": addr, })
                tx_params["gas"] = gas_estimate

                # Building, signing, and sending the approval transaction
                approval_contract = contract.functions.approve(exchange_con_addr, MAX_UINT256).build_transaction(tx_params)
                signed_tx = w3.eth.account.sign_transaction(approval_contract, self.secret_key)
                txn_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)
                reciept = w3.eth.wait_for_transaction_receipt(txn_hash)
                print(f"Approved allowance for token {token} with transaction hash {txn_hash}")
                receipts.append(reciept)
            except Exception as e:
                # Log the error and continue with the next token
                self.logger().debug("Error occurred while approving allowance for token %s: %s", token, str(e))
                if not fail_silently:
                    raise e
        return receipts if len(receipts) > 0 else None

    async def initialize_market_list(self):
        return await self._api_request(
            method=RESTMethod.GET,
            params={"page": 1, "sort_order": "desc", "sort_by": "volume", "page_size": 20, "verified": "true"},
            path_url = CONSTANTS.MARKET_LIST_PATH_URL.format(self.chain),
            is_auth_required = False,
            limit_id = CONSTANTS.MARKET_LIST_PATH_URL)

    async def initialize_verified_market(self):
        data = await self.initialize_market_list()
        id = []
        for trading_pair in self.trading_pairs:
            symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        for market in data:
            if market["chainId"] == self.chain and market["symbol"] == symbol:
                id.append(market)
        return await self._api_request(
            path_url = CONSTANTS.EXCHANGE_INFO_PATH_URL.format(self.chain, id[0]["id"]),
            method=RESTMethod.GET,
            is_auth_required = False,
            limit_id = CONSTANTS.EXCHANGE_INFO_PATH_URL)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        if symbol is not None:
            data = await self.initialize_verified_market()
            resp_json = await self._api_request(
                method=RESTMethod.GET,
                path_url = CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL.format(self.chain, data["id"]),
                is_auth_required = False,
                limit_id = CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL
            )
            return Decimal(resp_json["ticker"]["price"])

    async def _make_network_check_request(self):
        return await self._api_request(
            path_url = self.check_network_request_path,
            method=RESTMethod.GET,
            is_auth_required = False,
            limit_id = CONSTANTS.PING_PATH_URL)

    async def _make_trading_rules_request(self) -> Any:
        data: list[dict[str, Any]] = await self._api_request(
            path_url = self.trading_pairs_request_path.format(self.chain),
            method=RESTMethod.GET,
            params={"page": 1, "sort_order": "desc", "sort_by": "volume", "page_size": 20, "verified": "true"},
            is_auth_required = False,
            limit_id = CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL)
        return data

    async def _make_trading_pairs_request(self) -> Any:
        resp = await self._api_request(
            path_url = self.trading_pairs_request_path.format(self.chain),
            method=RESTMethod.GET,
            params={
                "page": 1,
                "sort_order": "desc",
                "sort_by": "volume",
                "page_size": 20,
                "verified": "true"
            },
            is_auth_required = False,
            limit_id = CONSTANTS.EXCHANGE_INFO_PATH_LIST_URL)
        return resp

    async def tokens_info(self):
        account_info = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.ACCOUNTS_PATH_URL.format(self.chain, self.api_key),
            limit_id=CONSTANTS.ACCOUNTS_PATH_URL,
            is_auth_required=False)
        data = []
        for dats in (account_info):
            token_data = {"symbol": dats["symbol"], "address": dats["address"]}
            data.append(token_data)
        return data



import logging
import time
from dataclasses import dataclass
from decimal import Decimal
from typing import Dict, NamedTuple, Optional, cast

from hummingbot.client.hummingbot_application import HummingbotApplication
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.logger.logger import HummingbotLogger


class GatewayPriceShimKey(NamedTuple):
    connector_name: str
    chain: str
    network: str
    trading_pair: str


@dataclass
class GatewayPriceShimEntry:
    from_exchange: str
    from_trading_pair: str
    to_connector_name: str
    to_chain: str
    to_network: str
    to_trading_pair: str


@dataclass
class GatewayPriceDeltaEntry:
    connector_name: str
    chain: str
    network: str
    trading_pair: str
    delta: Decimal
    end_timestamp: float


class GatewayPriceShim:
    """
    Developer / QA tool for modifying the apparent prices on DEX connectors (e.g. uniswap) during integration tests.

    When the gateway price shim is enabled for a particular DEX trading pair (e.g. "WETH-DAI" on uniswap), the apparent
    price on the DEX will follow the prices on a different trading pair on another exchange (e.g. "ETH-USDT" on
    Binance). The price shim then exposes a function (`apply_price_delta()`) which allows the developer to modify the
    apparent prices on the DEX trading pair during live trading. This means the developer can manually control and
    inject arbitrage opportunities in an amm_arb trading session, for carrying out integration tests.

    How to use:

    1. Set up an amm_arb strategy config that trades between a testnet trading pair (e.g. "WETH-DAI" on Uniswap Kovan),
       and a corresponding trading pair on a paper trading exchange (e.g. "ETH-USDT" on Binance paper trade).
    2. Set the `debug_price_shim` parameter of the strategy config to true.
    3. Start the strategy.
    4. Observe that the apparent AMM prices will follow the paper trading exchange prices.
    5. Inject price deltas to the AMM prices by going to the debug console, and issuing the following:

    ```
    from hummingbot.connector.gateway_price_shim import GatewayPriceShim
    from decimal import Decimal
    GatewayPriceShim.get_instance().apply_price_delta("uniswap", "ethereum", "kovan", "WETH-DAI", Decimal(40))
    ```

    6. Observe that the apparent AMM prices is increased by the delta amount, and the amm_arb strategy will start
       issuing arbitrage trades.
    7. Price delta values can be negative if you want to generate buy orders on the AMM side.

    """
    _gps_logger: Optional[HummingbotLogger] = None
    _shared_instance: Optional["GatewayPriceShim"] = None
    _shim_entries: Dict[GatewayPriceShimKey, GatewayPriceShimEntry]
    _delta_entries: Dict[GatewayPriceShimKey, GatewayPriceDeltaEntry]

    def __init__(self):
        self._shim_entries = {}
        self._delta_entries = {}

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._gps_logger is None:
            cls._gps_logger = cast(HummingbotLogger, logging.getLogger(__name__))
        return cls._gps_logger

    @classmethod
    def get_instance(cls) -> "GatewayPriceShim":
        if cls._shared_instance is None:
            cls._shared_instance = GatewayPriceShim()
        return cls._shared_instance

    def patch_prices(
            self,
            from_exchange: str,
            from_trading_pair: str,
            to_connector_name: str,
            to_chain: str,
            to_network: str,
            to_trading_pair: str
    ):
        key: GatewayPriceShimKey = GatewayPriceShimKey(
            connector_name=to_connector_name,
            chain=to_chain,
            network=to_network,
            trading_pair=to_trading_pair,
        )
        self._shim_entries[key] = GatewayPriceShimEntry(
            from_exchange=from_exchange,
            from_trading_pair=from_trading_pair,
            to_connector_name=to_connector_name,
            to_chain=to_chain,
            to_network=to_network,
            to_trading_pair=to_trading_pair
        )

    def apply_price_delta(
            self,
            connector_name: str,
            chain: str,
            network: str,
            trading_pair: str,
            delta: Decimal,
            duration_seconds: float = 60):
        key: GatewayPriceShimKey = GatewayPriceShimKey(
            connector_name=connector_name,
            chain=chain,
            network=network,
            trading_pair=trading_pair,
        )
        if key not in self._shim_entries:
            raise ValueError(f"The trading pair {trading_pair} on {chain}/{network}/{connector_name} has not had "
                             f"price shim installed yet.")
        self._delta_entries[key] = GatewayPriceDeltaEntry(
            connector_name=connector_name,
            chain=chain,
            network=network,
            trading_pair=trading_pair,
            delta=delta,
            end_timestamp=time.time() + duration_seconds
        )

    async def get_connector_price(
            self,
            connector_name: str,
            chain: str,
            network: str,
            trading_pair: str,
            is_buy: bool,
            amount: Decimal) -> Optional[Decimal]:
        key: GatewayPriceShimKey = GatewayPriceShimKey(
            connector_name=connector_name,
            chain=chain,
            network=network,
            trading_pair=trading_pair
        )
        if key not in self._shim_entries:
            return None

        shim_entry: GatewayPriceShimEntry = self._shim_entries[key]
        exchange_market: Optional[ExchangeBase] = HummingbotApplication.main_application().markets.get(
            shim_entry.from_exchange
        )
        if exchange_market is None:
            self.logger().warning(f"Gateway price shim failure: "
                                  f"reference exchange market '{shim_entry.from_exchange}' not found. "
                                  f"Going to use on-chain prices instead.")
            return None

        exchange_price: Decimal = await exchange_market.get_quote_price(
            shim_entry.from_trading_pair,
            is_buy,
            amount
        )
        if key in self._delta_entries:
            delta_entry: GatewayPriceDeltaEntry = self._delta_entries[key]
            now: float = time.time()
            if now <= delta_entry.end_timestamp:
                exchange_price += delta_entry.delta
            else:
                del self._delta_entries[key]
        return exchange_price

    def has_price_shim(self, connector_name: str, chain: str, network: str, trading_pair: str) -> bool:
        key: GatewayPriceShimKey = GatewayPriceShimKey(
            connector_name=connector_name,
            chain=chain,
            network=network,
            trading_pair=trading_pair
        )
        return key in self._shim_entries



import asyncio
import logging
import time
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_constants as constants
import hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_web_utils as web_utils
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_exchange import CoinbaseAdvancedTradeExchange

from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_order_book import (
    CoinbaseAdvancedTradeOrderBook,
)


class CoinbaseAdvancedTradeAPIOrderBookDataSource(OrderBookTrackerDataSource):
    HEARTBEAT_TIME_INTERVAL = 30.0
    TRADE_STREAM_ID = 1
    DIFF_STREAM_ID = 2
    ONE_HOUR = 60 * 60

    _logger: HummingbotLogger | logging.Logger | None = None

    @classmethod
    def logger(cls) -> HummingbotLogger | logging.Logger:
        if cls._logger is None:
            name: str = HummingbotLogger.logger_name_for_class(cls)
            cls._logger = logging.getLogger(name)
        return cls._logger

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'CoinbaseAdvancedTradeExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = constants.DEFAULT_DOMAIN):
        """
        Initialize the CoinbaseAdvancedTradeAPIUserStreamDataSource.

        :param trading_pairs: The list of trading pairs to subscribe to.
        :param connector: The CoinbaseAdvancedTradeExchangePairProtocol implementation.
        :param api_factory: The WebAssistantsFactory instance for creating the WSAssistant.
        :param domain: The domain for the WebSocket connection.
        """
        super().__init__(trading_pairs)
        self._domain: str = domain
        self._api_factory: WebAssistantsFactory = api_factory
        self._connector: 'CoinbaseAdvancedTradeExchange' = connector

        self._subscription_lock: Optional[asyncio.Lock] = None
        self._ws_assistant: Optional[WSAssistant] = None
        self._last_traded_prices: Dict[str, float] = defaultdict(lambda: 0.0)

        # Override the default base queue keys
        self._diff_messages_queue_key = constants.WS_ORDER_SUBSCRIPTION_CHANNELS.inverse["order_book_diff"]
        self._trade_messages_queue_key = constants.WS_ORDER_SUBSCRIPTION_CHANNELS.inverse["trade"]

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        # await asyncio.sleep(0)
        return {trading_pair: self._last_traded_prices[trading_pair] or 0.0 for trading_pair in trading_pairs}

    # Implemented methods

    async def _request_order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        params = {
            "product_id": await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        snapshot: Dict[str, Any] = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=constants.SNAPSHOT_EP, domain=self._domain),
            params=params,
            method=RESTMethod.GET,
            is_auth_required=True,
            throttler_limit_id=constants.SNAPSHOT_EP,
        )
        return snapshot

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the order book events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        https://docs.cdp.coinbase.com/advanced-trade/docs/ws-best-practices

        Recommended to use several subscriptions
        {
            "type": "subscribe",
            "product_ids": [
                "ETH-USD",
                "BTC-USD"
            ],
            "channel": "level2",

            # Complemented by the WSAssistant
            "signature": "XYZ",
            "api_key": "XXX",
            "timestamp": 1675974199
        }
        """
        try:
            symbols = []
            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                symbols.append(symbol)

            for channel in ["heartbeats", *constants.WS_ORDER_SUBSCRIPTION_KEYS]:
                payload = {
                    "type": "subscribe",
                    "product_ids": symbols,
                    "channel": channel,
                }
                await ws.send(WSJSONRequest(payload=payload, is_auth_required=True))

            self.logger().info(f"Subscribed to order book channels for: {', '.join(self._trading_pairs)}")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            self.logger().debug(f"Error: {e}")
            raise

    async def _connected_websocket_assistant(self) -> WSAssistant:
        self._ws_assistant: WSAssistant = await self._api_factory.get_ws_assistant()
        await self._ws_assistant.connect(ws_url=constants.WSS_URL.format(domain=self._domain), max_msg_size=constants.WS_MAX_MSG_SIZE)
        return self._ws_assistant

    # --- Implementation of abstract methods from the Base class ---
    # Unused methods
    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = time.time()
        snapshot_msg: OrderBookMessage = CoinbaseAdvancedTradeOrderBook.snapshot_message_from_exchange(
            snapshot,
            snapshot_timestamp,
            metadata={"trading_pair": trading_pair}
        )
        return snapshot_msg

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if raw_message is not None or "code" not in raw_message:
            event_type = raw_message["events"][0]["type"]
            if event_type == "update":
                events = raw_message["events"][0]
                trading_pair = events["trades"][0]["product_id"]
                # TODO: This code needs to be removed when Coinbase DIFF channel is fixed for USDC
                pair = await self.filter_pair(trading_pair)
                trade_message: OrderBookMessage = CoinbaseAdvancedTradeOrderBook.trade_message_from_exchange(
                    raw_message, {"trading_pair": pair})
                self.logger().debug(f"Order book message: {trade_message}")
                message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if raw_message is not None or "code" not in raw_message:
            event_type = raw_message["events"][0]["type"]
            if event_type == "update":
                trading_pair = raw_message["events"][0]["product_id"]
                # TODO: This code needs to be removed when Coinbase DIFF channel is fixed for USDC
                pair = await self.filter_pair(trading_pair)
                order_book_message: OrderBookMessage = CoinbaseAdvancedTradeOrderBook.diff_message_from_exchange(
                    raw_message, time.time(), {"trading_pair": pair})
                self.logger().debug(f"Order book message: {order_book_message}")
                message_queue.put_nowait(order_book_message)

    async def filter_pair(self, trading_pair):
        pairs = []
        for trad_pair in self._trading_pairs:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trad_pair)
            pairs.append(symbol)
        new_pair = ""
        base = trading_pair.split("-")[0]
        for symbol in pairs:
            symbol_base, symbol_quote = symbol.split("-")
            # Check if there's an exact match
            if symbol == trading_pair:
                new_pair = symbol
                break

            elif base == symbol_base and "USD" in symbol_quote:
                new_quote = "USDC"
                proposed_pair = f"{base}-{new_quote}"
                # Only update if the proposed pair exists in pairs
                if proposed_pair in pairs:
                    new_pair = proposed_pair
                    break

        return new_pair

    def _channel_originating_message(self, event_message: Dict[str, Any]):
        channel = ""
        if event_message and "channel" in event_message:
            if "events" in event_message:
                event_type = event_message.get("channel")
                if event_type in ["l2_data", "market_trades"]:
                    channel = (self._diff_messages_queue_key if event_type == constants.WS_ORDER_SUBSCRIPTION_CHANNELS.inverse["order_book_diff"]
                               else self._trade_messages_queue_key)
                return channel



import zlib
from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, SecretStr
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = True

EXAMPLE_PAIR = "ETH-USDT"

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.0025"),
    taker_percent_fee_decimal=Decimal("0.0025"),
)


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair
    :return: True if the trading pair is enabled, False otherwise
    """
    return exchange_info.get("trade_status", None) == "trading"


# Decompress WebSocket messages
def decompress_ws_message(message):
    if not isinstance(message, bytes):
        return message
    decompress = zlib.decompressobj(-zlib.MAX_WBITS)
    inflated = decompress.decompress(message)
    inflated += decompress.flush()
    return inflated.decode('UTF-8')


def compress_ws_message(message):
    if not isinstance(message, str):
        return message
    message = message.encode()
    compress = zlib.compressobj(wbits=-zlib.MAX_WBITS)
    deflated = compress.compress(message)
    deflated += compress.flush()
    return deflated


class BitmartConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="bitmart", client_data=None)
    bitmart_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your BitMart API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    bitmart_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your BitMart secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    bitmart_memo: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your BitMart API Memo",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="bitmart")


KEYS = BitmartConfigMap.construct()



from hummingbot.strategy.hedge.hedge import HedgeStrategy
from hummingbot.strategy.hedge.hedge_config_map_pydantic import MAX_CONNECTOR, HedgeConfigMap
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple


def start(self):
    c_map: HedgeConfigMap = self.strategy_config_map
    hedge_connector = c_map.hedge_connector.lower()
    hedge_markets = c_map.hedge_markets
    hedge_offsets = c_map.hedge_offsets
    offsets_dict = {hedge_connector: hedge_offsets}
    initialize_markets = [(hedge_connector, hedge_markets)]
    for i in range(MAX_CONNECTOR):
        connector_config = getattr(c_map, f"connector_{i}")
        connector = connector_config.connector
        if not connector:
            continue
        connector = connector.lower()
        markets = connector_config.markets
        offsets_dict[connector] = connector_config.offsets
        initialize_markets.append((connector, markets))
    self._initialize_markets(initialize_markets)
    self.market_trading_pair_tuples = []
    offsets_market_dict = {}
    for connector, markets in initialize_markets:
        offsets = offsets_dict[connector]
        for market, offset in zip(markets, offsets):
            base, quote = market.split("-")
            market_info = MarketTradingPairTuple(self.markets[connector], market, base, quote)
            self.market_trading_pair_tuples.append(market_info)
            offsets_market_dict[market_info] = offset
    index = len(hedge_markets)
    hedge_market_pairs = self.market_trading_pair_tuples[0:index]
    market_pairs = self.market_trading_pair_tuples[index:]
    self.strategy = HedgeStrategy(
        config_map=c_map, hedge_market_pairs=hedge_market_pairs, market_pairs=market_pairs, offsets=offsets_market_dict
    )



import asyncio
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from bidict import bidict

import hummingbot.connector.exchange.hashkey.hashkey_constants as CONSTANTS
import hummingbot.connector.exchange.hashkey.hashkey_utils as hashkey_utils
import hummingbot.connector.exchange.hashkey.hashkey_web_utils as web_utils
from hummingbot.connector.exchange.hashkey.hashkey_api_order_book_data_source import HashkeyAPIOrderBookDataSource
from hummingbot.connector.exchange.hashkey.hashkey_api_user_stream_data_source import HashkeyAPIUserStreamDataSource
from hummingbot.connector.exchange.hashkey.hashkey_auth import HashkeyAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter

s_logger = None
s_decimal_NaN = Decimal("nan")


class HashkeyExchange(ExchangePyBase):
    web_utils = web_utils

    def __init__(self,
                 client_config_map: "ClientConfigAdapter",
                 hashkey_api_key: str,
                 hashkey_api_secret: str,
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 ):
        self.api_key = hashkey_api_key
        self.secret_key = hashkey_api_secret
        self._domain = domain
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_trades_poll_hashkey_timestamp = 1.0
        super().__init__(client_config_map)

    @staticmethod
    def hashkey_order_type(order_type: OrderType) -> str:
        return order_type.name.upper()

    @staticmethod
    def to_hb_order_type(hashkey_type: str) -> OrderType:
        return OrderType[hashkey_type]

    @property
    def authenticator(self):
        return HashkeyAuth(
            api_key=self.api_key,
            secret_key=self.secret_key,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        if self._domain == "hashkey_global":
            return "hashkey"
        else:
            return self._domain

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.SERVER_TIME_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.MARKET, OrderType.LIMIT, OrderType.LIMIT_MAKER]

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_description = str(request_exception)
        is_time_synchronizer_related = ("-1021" in error_description
                                        and "Timestamp for the request" in error_description)
        return is_time_synchronizer_related

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return HashkeyAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            domain=self.domain,
            api_factory=self._web_assistants_factory,
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return HashkeyAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = order_type is OrderType.LIMIT_MAKER
        trade_base_fee = build_trade_fee(
            exchange=self.name,
            is_maker=is_maker,
            order_side=order_side,
            order_type=order_type,
            amount=amount,
            price=price,
            base_currency=base_currency,
            quote_currency=quote_currency
        )
        return trade_base_fee

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:
        amount_str = f"{amount:f}"
        type_str = self.hashkey_order_type(order_type)

        side_str = CONSTANTS.SIDE_BUY if trade_type is TradeType.BUY else CONSTANTS.SIDE_SELL
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        api_params = {"symbol": symbol,
                      "side": side_str,
                      "quantity": amount_str,
                      "type": type_str,
                      "recvWindow": 10000,
                      "newClientOrderId": order_id}
        path_url = CONSTANTS.ORDER_PATH_URL

        if order_type != OrderType.MARKET:
            api_params["price"] = f"{price:f}"
        else:
            path_url = CONSTANTS.MARKET_ORDER_PATH_URL

        if order_type == OrderType.LIMIT:
            api_params["timeInForce"] = CONSTANTS.TIME_IN_FORCE_GTC

        order_result = await self._api_post(
            path_url=path_url,
            params=api_params,
            is_auth_required=True,
            trading_pair=trading_pair,
            headers={"INPUT-SOURCE": CONSTANTS.HBOT_BROKER_ID},
        )

        o_id = str(order_result["orderId"])
        transact_time = int(order_result["transactTime"]) * 1e-3
        return (o_id, transact_time)

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        api_params = {}
        if tracked_order.exchange_order_id:
            api_params["orderId"] = tracked_order.exchange_order_id
        else:
            api_params["clientOrderId"] = tracked_order.client_order_id
        cancel_result = await self._api_delete(
            path_url=CONSTANTS.ORDER_PATH_URL,
            params=api_params,
            is_auth_required=True)

        if isinstance(cancel_result, dict) and "clientOrderId" in cancel_result:
            return True
        return False

    async def _format_trading_rules(self, exchange_info_dict: Dict[str, Any]) -> List[TradingRule]:
        """
        Example:
        {
            "timezone": "UTC",
            "serverTime": "1703696385826",
            "brokerFilters": [],
            "symbols": [
                {
                    "symbol": "ETHUSD",
                    "symbolName": "ETHUSD",
                    "status": "TRADING",
                    "baseAsset": "ETH",
                    "baseAssetName": "ETH",
                    "baseAssetPrecision": "0.0001",
                    "quoteAsset": "USD",
                    "quoteAssetName": "USD",
                    "quotePrecision": "0.0000001",
                    "retailAllowed": true,
                    "piAllowed": true,
                    "corporateAllowed": true,
                    "omnibusAllowed": true,
                    "icebergAllowed": false,
                    "isAggregate": false,
                    "allowMargin": false,
                    "filters": [
                        {
                            "minPrice": "0.01",
                            "maxPrice": "100000.00000000",
                            "tickSize": "0.01",
                            "filterType": "PRICE_FILTER"
                        },
                        {
                            "minQty": "0.005",
                            "maxQty": "53",
                            "stepSize": "0.0001",
                            "filterType": "LOT_SIZE"
                        },
                        {
                            "minNotional": "10",
                            "filterType": "MIN_NOTIONAL"
                        },
                        {
                            "minAmount": "10",
                            "maxAmount": "10000000",
                            "minBuyPrice": "0",
                            "filterType": "TRADE_AMOUNT"
                        },
                        {
                            "maxSellPrice": "0",
                            "buyPriceUpRate": "0.2",
                            "sellPriceDownRate": "0.2",
                            "filterType": "LIMIT_TRADING"
                        },
                        {
                            "buyPriceUpRate": "0.2",
                            "sellPriceDownRate": "0.2",
                            "filterType": "MARKET_TRADING"
                        },
                        {
                            "noAllowMarketStartTime": "0",
                            "noAllowMarketEndTime": "0",
                            "limitOrderStartTime": "0",
                            "limitOrderEndTime": "0",
                            "limitMinPrice": "0",
                            "limitMaxPrice": "0",
                            "filterType": "OPEN_QUOTE"
                        }
                    ]
                }
            ],
            "options": [],
            "contracts": [],
            "coins": [
                {
                    "orgId": "9001",
                    "coinId": "BTC",
                    "coinName": "BTC",
                    "coinFullName": "Bitcoin",
                    "allowWithdraw": true,
                    "allowDeposit": true,
                    "chainTypes": [
                        {
                            "chainType": "Bitcoin",
                            "withdrawFee": "0",
                            "minWithdrawQuantity": "0.0005",
                            "maxWithdrawQuantity": "0",
                            "minDepositQuantity": "0.0001",
                            "allowDeposit": true,
                            "allowWithdraw": true
                        }
                    ]
                },
                {
                    "orgId": "9001",
                    "coinId": "ETH",
                    "coinName": "ETH",
                    "coinFullName": "Ethereum",
                    "allowWithdraw": true,
                    "allowDeposit": true,
                    "chainTypes": [
                        {
                            "chainType": "ERC20",
                            "withdrawFee": "0",
                            "minWithdrawQuantity": "0",
                            "maxWithdrawQuantity": "0",
                            "minDepositQuantity": "0.0075",
                            "allowDeposit": true,
                            "allowWithdraw": true
                        }
                    ]
                },
                {
                    "orgId": "9001",
                    "coinId": "USD",
                    "coinName": "USD",
                    "coinFullName": "USD",
                    "allowWithdraw": true,
                    "allowDeposit": true,
                    "chainTypes": []
                }
            ]
        }
        """
        trading_pair_rules = exchange_info_dict.get("symbols", [])
        retval = []
        for rule in trading_pair_rules:
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule.get("symbol"))

                trading_filter_info = {item["filterType"]: item for item in rule.get("filters", [])}

                min_order_size = trading_filter_info.get("LOT_SIZE", {}).get("minQty")
                min_price_increment = trading_filter_info.get("PRICE_FILTER", {}).get("minPrice")
                min_base_amount_increment = rule.get("baseAssetPrecision")
                min_notional_size = trading_filter_info.get("TRADE_AMOUNT", {}).get("minAmount")

                retval.append(
                    TradingRule(trading_pair,
                                min_order_size=Decimal(min_order_size),
                                min_price_increment=Decimal(min_price_increment),
                                min_base_amount_increment=Decimal(min_base_amount_increment),
                                min_notional_size=Decimal(min_notional_size)))

            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {rule.get('symbol')}. Skipping.")
        return retval

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are balance updates, order updates and trade events.
        """
        async for event_messages in self._iter_user_event_queue():
            if isinstance(event_messages, dict) and "ping" in event_messages:
                continue

            for event_message in event_messages:
                try:
                    event_type = event_message.get("e")
                    if event_type == "executionReport":
                        execution_type = event_message.get("X")
                        client_order_id = event_message.get("c")
                        tracked_order = self._order_tracker.fetch_order(client_order_id=client_order_id)
                        if tracked_order is not None:
                            if execution_type in ["PARTIALLY_FILLED", "FILLED"]:
                                fee = TradeFeeBase.new_spot_fee(
                                    fee_schema=self.trade_fee_schema(),
                                    trade_type=tracked_order.trade_type,
                                    flat_fees=[TokenAmount(amount=Decimal(event_message["n"]), token=event_message["N"])]
                                )
                                trade_update = TradeUpdate(
                                    trade_id=str(event_message["d"]),
                                    client_order_id=client_order_id,
                                    exchange_order_id=str(event_message["i"]),
                                    trading_pair=tracked_order.trading_pair,
                                    fee=fee,
                                    fill_base_amount=Decimal(event_message["l"]),
                                    fill_quote_amount=Decimal(event_message["l"]) * Decimal(event_message["L"]),
                                    fill_price=Decimal(event_message["L"]),
                                    fill_timestamp=int(event_message["E"]) * 1e-3,
                                )
                                self._order_tracker.process_trade_update(trade_update)

                            order_update = OrderUpdate(
                                trading_pair=tracked_order.trading_pair,
                                update_timestamp=int(event_message["E"]) * 1e-3,
                                new_state=CONSTANTS.ORDER_STATE[event_message["X"]],
                                client_order_id=client_order_id,
                                exchange_order_id=str(event_message["i"]),
                            )
                            self._order_tracker.process_order_update(order_update=order_update)

                    elif event_type == "outboundAccountInfo":
                        balances = event_message["B"]
                        for balance_entry in balances:
                            asset_name = balance_entry["a"]
                            free_balance = Decimal(balance_entry["f"])
                            total_balance = Decimal(balance_entry["f"]) + Decimal(balance_entry["l"])
                            self._account_available_balances[asset_name] = free_balance
                            self._account_balances[asset_name] = total_balance

                except asyncio.CancelledError:
                    raise
                except Exception:
                    self.logger().error("Unexpected error in user stream listener loop.", exc_info=True)
                    await self._sleep(5.0)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = int(order.exchange_order_id)
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            fills_data = await self._api_get(
                path_url=CONSTANTS.MY_TRADES_PATH_URL,
                params={
                    "clientOrderId": order.client_order_id,
                },
                is_auth_required=True,
                limit_id=CONSTANTS.MY_TRADES_PATH_URL)
            if fills_data is not None:
                for trade in fills_data:
                    exchange_order_id = str(trade["orderId"])
                    if exchange_order_id != str(order.exchange_order_id):
                        continue
                    fee = TradeFeeBase.new_spot_fee(
                        fee_schema=self.trade_fee_schema(),
                        trade_type=order.trade_type,
                        percent_token=trade["commissionAsset"],
                        flat_fees=[TokenAmount(amount=Decimal(trade["commission"]), token=trade["commissionAsset"])]
                    )
                    trade_update = TradeUpdate(
                        trade_id=str(trade["ticketId"]),
                        client_order_id=order.client_order_id,
                        exchange_order_id=exchange_order_id,
                        trading_pair=trading_pair,
                        fee=fee,
                        fill_base_amount=Decimal(trade["qty"]),
                        fill_quote_amount=Decimal(trade["price"]) * Decimal(trade["qty"]),
                        fill_price=Decimal(trade["price"]),
                        fill_timestamp=int(trade["time"]) * 1e-3,
                    )
                    trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        updated_order_data = await self._api_get(
            path_url=CONSTANTS.ORDER_PATH_URL,
            params={
                "origClientOrderId": tracked_order.client_order_id},
            is_auth_required=True)

        new_state = CONSTANTS.ORDER_STATE[updated_order_data["status"]]

        order_update = OrderUpdate(
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=str(updated_order_data["orderId"]),
            trading_pair=tracked_order.trading_pair,
            update_timestamp=int(updated_order_data["updateTime"]) * 1e-3,
            new_state=new_state,
        )

        return order_update

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        account_info = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.ACCOUNTS_PATH_URL,
            is_auth_required=True)
        balances = account_info["balances"]
        for balance_entry in balances:
            asset_name = balance_entry["asset"]
            free_balance = Decimal(balance_entry["free"])
            total_balance = Decimal(balance_entry["total"])
            self._account_available_balances[asset_name] = free_balance
            self._account_balances[asset_name] = total_balance
            remote_asset_names.add(asset_name)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(hashkey_utils.is_exchange_information_valid, exchange_info["symbols"]):
            mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(base=symbol_data["baseAsset"],
                                                                        quote=symbol_data["quoteAsset"])
        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
        }
        resp_json = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.LAST_TRADED_PRICE_PATH,
            params=params,
        )

        return float(resp_json["price"])

    async def _api_request(self,
                           path_url,
                           method: RESTMethod = RESTMethod.GET,
                           params: Optional[Dict[str, Any]] = None,
                           data: Optional[Dict[str, Any]] = None,
                           is_auth_required: bool = False,
                           return_err: bool = False,
                           limit_id: Optional[str] = None,
                           trading_pair: Optional[str] = None,
                           **kwargs) -> Dict[str, Any]:
        last_exception = None
        rest_assistant = await self._web_assistants_factory.get_rest_assistant()
        url = web_utils.rest_url(path_url, domain=self.domain)
        local_headers = {
            "Content-Type": "application/x-www-form-urlencoded"}
        for _ in range(2):
            try:
                request_result = await rest_assistant.execute_request(
                    url=url,
                    params=params,
                    data=data,
                    method=method,
                    is_auth_required=is_auth_required,
                    return_err=return_err,
                    headers=local_headers,
                    throttler_limit_id=limit_id if limit_id else path_url,
                )
                return request_result
            except IOError as request_exception:
                last_exception = request_exception
                if self._is_request_exception_related_to_time_synchronizer(request_exception=request_exception):
                    self._time_synchronizer.clear_time_offset_ms_samples()
                    await self._update_time_synchronizer()
                else:
                    raise

        # Failed even after the last retry
        raise last_exception



from typing import Dict, Optional

from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType


class TegroOrderBook(OrderBook):

    @classmethod
    def snapshot_message_from_exchange(cls, msg: Dict[str, any], timestamp: Optional[float] = None,
                                       metadata: Optional[Dict] = None) -> OrderBookMessage:
        if metadata:
            msg.update(metadata)

        def accumulate_quantities(entries, reverse=False):
            cumulative_quantity = 0.0
            cumulative_data = []

            # If reverse is True, process from the lowest price to the highest
            if entries is not None:
                entries = entries[::-1] if reverse else entries

                for entry in entries:
                    price = float(entry['price'])  # Keep price unchanged
                    quantity = float(entry['quantity'])
                    cumulative_quantity += quantity  # Only accumulate the quantity
                    cumulative_data.append([price, cumulative_quantity])  # Price remains the same

                # Reverse again if asks were reversed to maintain order in the result
                return cumulative_data[::-1] if reverse else cumulative_data
            else:
                return cumulative_data
        # For asks, reverse the order of accumulation (because lower prices come first)
        asks = accumulate_quantities(msg.get('asks', []), reverse=True)
        # For bids, accumulate as usual
        bids = accumulate_quantities(msg.get('bids', []), reverse=False)

        return OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": msg["trading_pair"],
            "update_id": msg["timestamp"],
            "bids": bids,
            "asks": asks
        }, timestamp=timestamp)

    @classmethod
    def diff_message_from_exchange(cls,
                                   msg: Dict[str, any],
                                   timestamp: Optional[float] = None,
                                   metadata: Optional[Dict] = None) -> OrderBookMessage:
        """
        Creates a diff message with the changes in the order book received from the exchange
        :param msg: the changes in the order book
        :param timestamp: the timestamp of the difference
        :param metadata: a dictionary with extra information to add to the difference data
        :return: a diff message with the changes in the order book notified by the exchange
        """
        if metadata:
            msg.update(metadata)

        def accumulate_quantities(entries, reverse=False):
            cumulative_quantity = 0.0
            cumulative_data = []

            # If reverse is True, process from the lowest price to the highest
            entries = entries[::-1] if reverse else entries

            for entry in entries:
                price = float(entry['price'])  # Keep price unchanged
                quantity = float(entry['quantity'])
                cumulative_quantity += quantity  # Only accumulate the quantity
                cumulative_data.append([price, cumulative_quantity])  # Price remains the same

            # Reverse again if asks were reversed to maintain order in the result
            return cumulative_data[::-1] if reverse else cumulative_data

        # For asks, reverse the order of accumulation (because lower prices come first)
        asks = accumulate_quantities(msg["data"].get('asks', []), reverse=True)
        # For bids, accumulate as usual
        bids = accumulate_quantities(msg["data"].get('bids', []), reverse=False)

        return OrderBookMessage(OrderBookMessageType.DIFF, {
            "trading_pair": msg["trading_pair"],
            "update_id": msg["data"]["timestamp"],
            "bids": bids,
            "asks": asks
        }, timestamp=timestamp * 1e-3)

    @classmethod
    def trade_message_from_exchange(cls, msg: Dict[str, any], timestamp: Optional[float] = None, metadata: Optional[Dict] = None):
        """
        Creates a trade message with the information from the trade event sent by the exchange
        :param msg: the trade event details sent by the exchange
        :param metadata: a dictionary with extra information to add to trade message
        :return: a trade message with the details of the trade as provided by the exchange
        """
        if metadata:
            msg.update(metadata)
        ts = timestamp
        return OrderBookMessage(OrderBookMessageType.TRADE, {
            "trading_pair": msg["data"]["symbol"],
            "trade_type": float(TradeType.BUY.value) if msg["data"]["is_buyer_maker"] else float(TradeType.SELL.value),
            "trade_id": msg["data"]["id"],
            "update_id": ts,
            "price": msg["data"]["price"],
            "amount": msg["data"]["amount"]
        }, timestamp=ts * 1e-3)



import asyncio
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.tegro import tegro_constants as CONSTANTS, tegro_web_utils
from hummingbot.connector.exchange.tegro.tegro_order_book import TegroOrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.tegro.tegro_exchange import TegroExchange


class TegroAPIOrderBookDataSource(OrderBookTrackerDataSource):
    FULL_ORDER_BOOK_RESET_DELTA_SECONDS = 2
    HEARTBEAT_TIME_INTERVAL = 30.0
    TRADE_STREAM_ID = 1
    DIFF_STREAM_ID = 2
    ONE_HOUR = 60 * 60

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'TegroExchange',
                 api_factory: WebAssistantsFactory,
                 domain: Optional[str] = CONSTANTS.DOMAIN):
        super().__init__(trading_pairs)
        self._connector = connector
        self.trading_pairs = trading_pairs
        self._trade_messages_queue_key = CONSTANTS.TRADE_EVENT_TYPE
        self._diff_messages_queue_key = CONSTANTS.DIFF_EVENT_TYPE
        self._domain: Optional[str] = domain
        self._api_factory = api_factory

    @property
    def chain_id(self):
        return self._connector._chain

    @property
    def chain(self):
        chain = 8453
        if self._domain.endswith("_testnet"):
            chain = CONSTANTS.TESTNET_CHAIN_IDS[self.chain_id]
        elif self._domain == "tegro":
            chain_id = CONSTANTS.DEFAULT_CHAIN
            # In this case tegro is default to base mainnet
            chain = CONSTANTS.MAINNET_CHAIN_IDS[chain_id]
        return chain

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        data = await self.initialize_verified_market()
        params = {
            "chain_id": self.chain,
            "market_id": data["id"],
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=tegro_web_utils.public_rest_url(CONSTANTS.SNAPSHOT_PATH_URL, domain=self._domain),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.SNAPSHOT_PATH_URL,
        )
        return data

    async def initialize_verified_market(self):
        data = await self.initialize_market_list()
        id = []
        for trading_pair in self._trading_pairs:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        for market in data:
            if market["chainId"] == self.chain and market["symbol"] == symbol:
                id.append(market)
        rest_assistant = await self._api_factory.get_rest_assistant()
        return await rest_assistant.execute_request(
            url = tegro_web_utils.public_rest_url(CONSTANTS.EXCHANGE_INFO_PATH_URL.format(
                self.chain, id[0]["id"]), self._domain),
            method=RESTMethod.GET,
            is_auth_required = False,
            throttler_limit_id = CONSTANTS.EXCHANGE_INFO_PATH_URL,
        )

    async def initialize_market_list(self):
        rest_assistant = await self._api_factory.get_rest_assistant()
        return await rest_assistant.execute_request(
            method=RESTMethod.GET,
            params={
                "page": 1,
                "sort_order": "desc",
                "page_size": 20,
                "verified": "true"
            },
            url = tegro_web_utils.public_rest_url(CONSTANTS.MARKET_LIST_PATH_URL.format(self.chain), self._domain),
            is_auth_required = False,
            throttler_limit_id = CONSTANTS.MARKET_LIST_PATH_URL,
        )

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            market_data = await self.initialize_market_list()
            param: str = self._process_market_data(market_data)

            payload = {
                "action": "subscribe",
                "channelId": param
            }
            subscribe_request: WSJSONRequest = WSJSONRequest(payload=payload)
            await ws.send(subscribe_request)

            self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            raise

    def _process_market_data(self, market_data):
        symbol = ""
        for market in market_data:
            s = market["symbol"]
            symb = s.split("_")
            new_symbol = f"{symb[0]}-{symb[1]}"
            if new_symbol in self._trading_pairs:
                address = str(market["base_contract_address"])
                symbol = f"{self.chain}/{address}"
                break
        return symbol

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=tegro_web_utils.wss_url(CONSTANTS.PUBLIC_WS_ENDPOINT, self._domain),
                         ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = time.time()
        snapshot_msg: OrderBookMessage = TegroOrderBook.snapshot_message_from_exchange(
            snapshot,
            snapshot_timestamp,
            metadata={"trading_pair": trading_pair}
        )
        return snapshot_msg

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if "result" not in raw_message:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=raw_message["data"]["symbol"])
            trade_message = TegroOrderBook.trade_message_from_exchange(
                raw_message, time.time(), {"trading_pair": trading_pair})
            message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if "result" not in raw_message:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=raw_message["data"]["symbol"])
            order_book_message: OrderBookMessage = TegroOrderBook.diff_message_from_exchange(
                raw_message, time.time(), {"trading_pair": trading_pair})
            message_queue.put_nowait(order_book_message)
        return

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "action" in event_message:
            event_channel = event_message.get("action")
            if event_channel == CONSTANTS.TRADE_EVENT_TYPE:
                channel = self._trade_messages_queue_key
            if event_channel == CONSTANTS.DIFF_EVENT_TYPE:
                channel = self._diff_messages_queue_key
        return channel



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import OrderState

MAX_ID_BIT_COUNT = 63  # experimentally, 64 bit ints sometimes result in OMS assigning order IDs of zero
MAX_ORDER_NOT_FOUND_ON_CANCEL = 2

# rest endpoints
REST_AUTH_ENDPOINT = "Authenticate"
REST_PRODUCTS_ENDPOINT = "GetInstruments"
REST_GET_L1_ENDPOINT = "GetLevel1"
REST_GET_L2_SNAPSHOT_ENDPOINT = "GetL2Snapshot"
REST_PING_ENDPOINT = "Ping"
REST_ORDER_CREATION_ENDPOINT = "SendOrder"
REST_ORDER_STATUS_ENDPOINT = "GetOrderStatus"
REST_ORDER_CANCELATION_ENDPOINT = "CancelOrder"
REST_ACC_POSITIONS_ENDPOINT = "GetAccountPositions"
REST_TRADE_HISTORY_ENDPOINT = "GetTradesHistory"
_ALL_REST_ENDPOINTS = [
    REST_AUTH_ENDPOINT,
    REST_PRODUCTS_ENDPOINT,
    REST_GET_L1_ENDPOINT,
    REST_GET_L2_SNAPSHOT_ENDPOINT,
    REST_PING_ENDPOINT,
    REST_ORDER_CREATION_ENDPOINT,
    REST_ORDER_STATUS_ENDPOINT,
    REST_ORDER_CANCELATION_ENDPOINT,
    REST_ACC_POSITIONS_ENDPOINT,
    REST_TRADE_HISTORY_ENDPOINT,
]

# ws endpoints
WS_AUTH_ENDPOINT = "AuthenticateUser"
WS_ACC_EVENTS_ENDPOINT = "SubscribeAccountEvents"
WS_TRADES_SUB_ENDPOINT = "SubscribeTrades"
WS_L2_SUB_ENDPOINT = "SubscribeLevel2"
WS_PING_REQUEST = "Ping"
_ALL_WS_ENDPOINTS = [
    WS_AUTH_ENDPOINT,
    WS_ACC_EVENTS_ENDPOINT,
    WS_TRADES_SUB_ENDPOINT,
    WS_L2_SUB_ENDPOINT,
    WS_PING_REQUEST,
]

# ws events
WS_L2_EVENT = "Level2UpdateEvent"
WS_ACC_POS_EVENT = "AccountPositionEvent"
WS_ORDER_STATE_EVENT = "OrderStateEvent"
WS_ORDER_TRADE_EVENT = "OrderTradeEvent"
WS_CANCEL_ORDER_REJECTED_EVENT = "CancelOrderRejectEvent"

# limits
REST_REQ_LIMIT_ID = "WSReqLimitID"
REST_REQ_LIMIT = 5_000
WS_REQ_LIMIT_ID = "WSReqLimitID"
WS_REQ_LIMIT = 500_000

RATE_LIMITS = [
    RateLimit(REST_AUTH_ENDPOINT, limit=5_000, time_interval=60),
    RateLimit(REST_REQ_LIMIT_ID, limit=REST_REQ_LIMIT, time_interval=60),
    RateLimit(WS_AUTH_ENDPOINT, limit=50_000, time_interval=60),
    RateLimit(WS_REQ_LIMIT_ID, limit=WS_REQ_LIMIT, time_interval=60),
]
for e in _ALL_REST_ENDPOINTS:
    RATE_LIMITS.append(  # each limit defined separately so that children can be more granular
        RateLimit(
            limit_id=e,
            limit=REST_REQ_LIMIT,
            time_interval=60,
            linked_limits=[LinkedLimitWeightPair(limit_id=REST_REQ_LIMIT_ID)],
        )
    )
for e in _ALL_WS_ENDPOINTS:  # noqa: F821
    RATE_LIMITS.append(  # each limit defined separately so that children can be more granular
        RateLimit(
            limit_id=e,
            limit=WS_REQ_LIMIT,
            time_interval=60,
            linked_limits=[LinkedLimitWeightPair(limit_id=WS_REQ_LIMIT_ID)]
        )
    )

# endpoint constant settings
MAX_L2_SNAPSHOT_DEPTH = 400
INCLUDE_LAST_COUNT = 0

# msg types
REQ_MSG_TYPE = 0
RESP_MSG_TYPE = 1
EVENT_MSG_TYPE = 3
ERROR_MSG_TYPE = 5

# time in force types
GTC_TIF = 1

# order types
LIMIT_ORDER_TYPE = 2
ORDER_TYPES = {
    OrderType.LIMIT: LIMIT_ORDER_TYPE
}

# order actions
BUY_ACTION = 0
SELL_ACTION = 1
ORDER_SIDE_MAP = {
    "Buy": TradeType.BUY,
    "Sell": TradeType.SELL,
}

# order state
ACTIVE_ORDER_STATE = "Working"  # can be either OPEN or PARTIALLY_FILLED
CANCELED_ORDER_STATE = "Canceled"
REJECTED_ORDER_STATE = "Rejected"
EXPIRED_ORDER_STATE = "Expired"
FULLY_EXECUTED_ORDER_STATE = "FullyExecuted"
ORDER_STATE_MAP = {
    CANCELED_ORDER_STATE: OrderState.CANCELED,
    REJECTED_ORDER_STATE: OrderState.FAILED,
    EXPIRED_ORDER_STATE: OrderState.FAILED,
    FULLY_EXECUTED_ORDER_STATE: OrderState.FILLED,
}

# fields
OMS_ID_FIELD = "OMSId"
USER_FIELD = "User"
USER_ID_FIELD = "UserId"
USER_NAME_FIELD = "UserName"
ACCOUNT_ID_FIELD = "AccountId"
INSTRUMENT_ID_FIELD = "InstrumentId"
BASE_FIELD = "Product1Symbol"
BASE_ID_FIELD = "Product1"
QUOTE_FIELD = "Product2Symbol"
QUOTE_ID_FIELD = "Product2"
FEE_PRODUCT_ID_FIELD = "FeeProductId"
FEE_AMOUNT_FIELD = "Fee"
START_TIME_FIELD = "StartTime"
TRADE_ID_FIELD = "TradeId"
AUTHENTICATED_FIELD = "Authenticated"
SESSION_TOKEN_FIELD = "SessionToken"
API_KEY_FIELD = "APIKey"
SIGNATURE_FIELD = "Signature"
NONCE_FIELD = "Nonce"
DEPTH_FIELD = "Depth"
INCLUDE_LAST_COUNT_FIELD = "IncludeLastCount"
TIME_IN_FORCE_FIELD = "TimeInForce"
CLIENT_ORDER_ID_FIELD = "ClientOrderId"
CL_ORDER_ID_FIELD = "ClOrderId"  # yes, this and the above are not typos...
ORDER_ID_FIELD = "OrderId"
SIDE_FIELD = "Side"
QUANTITY_FIELD = "quantity"
ORDER_TYPE_FIELD = "OrderType"
LIMIT_PRICE_FIELD = "LimitPrice"
PRICE_FIELD = "Price"
TRADE_TIME_MS_FIELD = "TradeTimeMS"
RESULT_FIELD = "result"
ERROR_CODE_FIELD = "errorcode"
ERROR_MSG_FIELD = "errormsg"
PRODUCT_SYMBOL_FIELD = "ProductSymbol"
SYMBOL_FIELD = "Symbol"
AMOUNT_FIELD = "Amount"
ORIGINAL_QUANTITY_FIELD = "OrigQuantity"
QUANTITY_EXECUTED_FIELD = "QuantityExecuted"
ORDER_STATE_FIELD = "OrderState"
AMOUNT_ON_HOLD_FIELD = "Hold"
ORDER_UPDATE_TS_FIELD = "LastUpdatedTime"
IS_DISABLED_FIELD = "IsDisable"
SESSION_STATUS_FIELD = "SessionStatus"
MIN_QUANT_FIELD = "MinimumQuantity"
MIN_PRICE_INCR_FIELD = "PriceIncrement"
MIN_QUANT_INCR_FIELD = "QuantityIncrement"
RECEIVE_TIME_FIELD = "ReceiveTime"
LAST_TRADED_PRICE_FIELD = "LastTradedPx"
MSG_TYPE_FIELD = "m"
MSG_SEQUENCE_FIELD = "i"
MSG_ENDPOINT_FIELD = "n"
MSG_DATA_FIELD = "o"

TRADE_UPDATE_INSTRUMENT_ID_FIELD = 1
TRADE_UPDATE_AMOUNT_FIELD = 2
TRADE_UPDATE_PRICE_FIELD = 3
TRADE_UPDATE_TS_FIELD = 6
TRADE_UPDATE_SIDE_FIELD = 8

DIFF_UPDATE_TS_FIELD = 2
DIFF_UPDATE_PRICE_FIELD = 6
DIFF_UPDATE_INSTRUMENT_ID_FIELD = 7
DIFF_UPDATE_AMOUNT_FIELD = 8
DIFF_UPDATE_SIDE_FIELD = 9

# other
RESOURCE_NOT_FOUND_ERR_CODE = 104
WS_MESSAGE_TIMEOUT = 20



import importlib
import inspect
import os
import re
import sys
from os import listdir
from os.path import exists, isfile, join
from typing import List

from prompt_toolkit.completion import CompleteEvent, Completer, WordCompleter
from prompt_toolkit.document import Document

from hummingbot.client import settings
from hummingbot.client.command.connect_command import OPTIONS as CONNECT_OPTIONS
from hummingbot.client.config.config_data_types import BaseClientModel
from hummingbot.client.settings import (
    GATEWAY_CONNECTORS,
    SCRIPT_STRATEGIES_PATH,
    SCRIPT_STRATEGY_CONF_DIR_PATH,
    STRATEGIES,
    STRATEGIES_CONF_DIR_PATH,
    AllConnectorSettings,
)
from hummingbot.client.ui.parser import ThrowingArgumentParser
from hummingbot.core.rate_oracle.rate_oracle import RATE_ORACLE_SOURCES
from hummingbot.core.utils.gateway_config_utils import list_gateway_wallets
from hummingbot.core.utils.trading_pair_fetcher import TradingPairFetcher
from hummingbot.strategy.strategy_v2_base import StrategyV2ConfigBase


def file_name_list(path, file_extension):
    if not exists(path):
        return []
    return sorted([f for f in listdir(path) if isfile(join(path, f)) and f.endswith(file_extension)])


class HummingbotCompleter(Completer):
    def __init__(self, hummingbot_application):
        super(HummingbotCompleter, self).__init__()
        self.hummingbot_application = hummingbot_application
        self._path_completer = WordCompleter(file_name_list(str(STRATEGIES_CONF_DIR_PATH), "yml"))
        self._command_completer = WordCompleter(self.parser.commands, ignore_case=True)
        self._exchange_completer = WordCompleter(sorted(AllConnectorSettings.get_connector_settings().keys()), ignore_case=True)
        self._spot_exchange_completer = WordCompleter(sorted(AllConnectorSettings.get_exchange_names()), ignore_case=True)
        self._exchange_amm_completer = WordCompleter(sorted(AllConnectorSettings.get_gateway_amm_connector_names()), ignore_case=True)
        self._exchange_clob_completer = WordCompleter(sorted(AllConnectorSettings.get_exchange_names()), ignore_case=True)
        self._trading_timeframe_completer = WordCompleter(["infinite", "from_date_to_date", "daily_between_times"], ignore_case=True)
        self._derivative_completer = WordCompleter(AllConnectorSettings.get_derivative_names(), ignore_case=True)
        self._derivative_exchange_completer = WordCompleter(AllConnectorSettings.get_derivative_names(), ignore_case=True)
        self._connect_option_completer = WordCompleter(CONNECT_OPTIONS, ignore_case=True)
        self._export_completer = WordCompleter(["keys", "trades"], ignore_case=True)
        self._balance_completer = WordCompleter(["limit", "paper"], ignore_case=True)
        self._history_completer = WordCompleter(["--days", "--verbose", "--precision"], ignore_case=True)
        self._gateway_completer = WordCompleter(["balance", "config", "connect", "connector-tokens", "generate-certs", "test-connection", "list", "approve-tokens"], ignore_case=True)
        self._gateway_connect_completer = WordCompleter(GATEWAY_CONNECTORS, ignore_case=True)
        self._gateway_connector_tokens_completer = self._exchange_amm_completer
        self._gateway_balance_completer = self._exchange_amm_completer
        self._gateway_approve_tokens_completer = self._exchange_amm_completer
        self._gateway_config_completer = WordCompleter(hummingbot_application.gateway_config_keys, ignore_case=True)
        self._strategy_completer = WordCompleter(STRATEGIES, ignore_case=True)
        self._script_strategy_completer = WordCompleter(file_name_list(str(SCRIPT_STRATEGIES_PATH), "py"))
        self._scripts_config_completer = WordCompleter(file_name_list(str(SCRIPT_STRATEGY_CONF_DIR_PATH), "yml"))
        self._strategy_v2_create_config_completer = self.get_strategies_v2_with_config()
        self._controller_completer = self.get_available_controllers()
        self._rate_oracle_completer = WordCompleter(list(RATE_ORACLE_SOURCES.keys()), ignore_case=True)
        self._mqtt_completer = WordCompleter(["start", "stop", "restart"], ignore_case=True)
        self._gateway_chains = []
        self._gateway_networks = []
        self._list_gateway_wallets_parameters = {"wallets": [], "chain": ""}

    def get_strategies_v2_with_config(self):
        file_names = file_name_list(str(SCRIPT_STRATEGIES_PATH), "py")
        strategies_with_config = []

        for script_name in file_names:
            try:
                script_name = script_name.replace(".py", "")
                module = sys.modules.get(f"{settings.SCRIPT_STRATEGIES_MODULE}.{script_name}")
                if module is not None:
                    script_module = importlib.reload(module)
                else:
                    script_module = importlib.import_module(f".{script_name}",
                                                            package=settings.SCRIPT_STRATEGIES_MODULE)
                config_class = next((member for member_name, member in inspect.getmembers(script_module)
                                     if inspect.isclass(member) and member not in [BaseClientModel, StrategyV2ConfigBase] and
                                     (issubclass(member, BaseClientModel) or issubclass(member, StrategyV2ConfigBase))))
                if config_class:
                    strategies_with_config.append(script_name)
            except Exception:
                pass
        return WordCompleter(strategies_with_config, ignore_case=True)

    def get_available_controllers(self):
        controllers_path = settings.CONTROLLERS_PATH  # Ensure this points to the controllers directory
        available_controllers = []

        for root, dirs, files in os.walk(controllers_path):
            for file in files:
                if file.endswith(".py") and not file.startswith("__"):
                    # Extract controller name from file path
                    relative_path = os.path.relpath(os.path.join(root, file), controllers_path)
                    controller_name = os.path.splitext(relative_path)[0].replace(os.path.sep, ".")
                    available_controllers.append(controller_name)

        return WordCompleter(available_controllers, ignore_case=True)

    def set_gateway_chains(self, gateway_chains):
        self._gateway_chains = gateway_chains

    def set_gateway_networks(self, gateway_networks):
        self._gateway_networks = gateway_networks

    def set_list_gateway_wallets_parameters(self, wallets, chain):
        self._list_gateway_wallets_parameters = {"wallets": wallets, "chain": chain}

    @property
    def prompt_text(self) -> str:
        return self.hummingbot_application.app.prompt_text

    @property
    def parser(self) -> ThrowingArgumentParser:
        return self.hummingbot_application.parser

    def get_subcommand_completer(self, first_word: str) -> Completer:
        subcommands: List[str] = self.parser.subcommands_from(first_word)
        return WordCompleter(subcommands, ignore_case=True)

    @property
    def _trading_pair_completer(self) -> Completer:
        trading_pair_fetcher = TradingPairFetcher.get_instance()
        market = ""
        for exchange in sorted(list(AllConnectorSettings.get_connector_settings().keys()), key=len, reverse=True):
            if exchange in self.prompt_text:
                market = exchange
                break
        trading_pairs = trading_pair_fetcher.trading_pairs.get(market, []) if trading_pair_fetcher.ready and market else []
        return WordCompleter(trading_pairs, ignore_case=True, sentence=True)

    @property
    def _gateway_chain_completer(self):
        return WordCompleter(self._gateway_chains, ignore_case=True)

    @property
    def _gateway_network_completer(self):
        return WordCompleter(self._gateway_networks, ignore_case=True)

    @property
    def _gateway_wallet_address_completer(self):
        return WordCompleter(list_gateway_wallets(self._list_gateway_wallets_parameters["wallets"], self._list_gateway_wallets_parameters["chain"]), ignore_case=True)

    @property
    def _option_completer(self):
        outer = re.compile(r"\((.+)\)")
        inner_str = outer.search(self.prompt_text).group(1)
        options = inner_str.split("/") if "/" in inner_str else []
        return WordCompleter(options, ignore_case=True)

    @property
    def _config_completer(self):
        config_keys = self.hummingbot_application.configurable_keys()
        return WordCompleter(config_keys, ignore_case=True)

    def _complete_strategies(self, document: Document) -> bool:
        return "strategy" in self.prompt_text and "strategy file" not in self.prompt_text

    def _complete_pmm_script_files(self, document: Document) -> bool:
        return "PMM script file" in self.prompt_text

    def _complete_configs(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("config")

    def _complete_options(self, document: Document) -> bool:
        return "(" in self.prompt_text and ")" in self.prompt_text and "/" in self.prompt_text

    def _complete_exchanges(self, document: Document) -> bool:
        return any(x for x in ("exchange name", "name of exchange", "name of the exchange")
                   if x in self.prompt_text.lower())

    def _complete_derivatives(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return "perpetual" in text_before_cursor or \
               any(x for x in ("derivative connector", "derivative name", "name of derivative", "name of the derivative")
                   if x in self.prompt_text.lower())

    def _complete_connect_options(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("connect ")

    def _complete_exchange_amm_connectors(self, document: Document) -> bool:
        return "(Exchange/AMM/CLOB)" in self.prompt_text

    def _complete_exchange_clob_connectors(self, document: Document) -> bool:
        return "(Exchange/AMM/CLOB)" in self.prompt_text

    def _complete_spot_exchanges(self, document: Document) -> bool:
        return "spot" in self.prompt_text

    def _complete_trading_timeframe(self, document: Document) -> bool:
        return any(x for x in ("trading timeframe", "execution timeframe")
                   if x in self.prompt_text.lower())

    def _complete_export_options(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return "export" in text_before_cursor

    def _complete_balance_options(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("balance ")

    def _complete_history_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("history ")

    def _complete_gateway_connect_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("gateway connect ")

    def _complete_gateway_connector_tokens_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("gateway connector-tokens ")

    def _complete_gateway_balance_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("gateway balance ")

    def _complete_gateway_approve_tokens_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("gateway approve-tokens ")

    def _complete_gateway_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("gateway ") and not text_before_cursor.startswith("gateway config ")

    def _complete_gateway_config_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("gateway config ")

    def _complete_script_strategy_files(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("start --script ") and "--conf" not in text_before_cursor

    def _complete_script_strategy_config(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("start --script ") and "--conf" in text_before_cursor

    def _complete_strategy_v2_files_with_config(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("create --script-config ")

    def _complete_controllers_config(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("create --controller-config ")

    def _complete_trading_pairs(self, document: Document) -> bool:
        return "trading pair" in self.prompt_text

    def _complete_paths(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return (("path" in self.prompt_text and "file" in self.prompt_text) or
                "import" in text_before_cursor)

    def _complete_gateway_chain(self, document: Document) -> bool:
        return "Which chain do you want" in self.prompt_text

    def _complete_gateway_network(self, document: Document) -> bool:
        return "Which network do you want" in self.prompt_text

    def _complete_gateway_wallet_addresses(self, document: Document) -> bool:
        return "Select a gateway wallet" in self.prompt_text

    def _complete_command(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return " " not in text_before_cursor and len(self.prompt_text.replace(">>> ", "")) == 0

    def _complete_subcommand(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        index: int = text_before_cursor.index(' ')
        return text_before_cursor[0:index] in self.parser.commands

    def _complete_balance_limit_exchanges(self, document: Document):
        text_before_cursor: str = document.text_before_cursor
        command_args = text_before_cursor.split(" ")
        return len(command_args) == 3 and command_args[0] == "balance" and command_args[1] == "limit"

    def _complete_rate_oracle_source(self, document: Document):
        return all(x in self.prompt_text for x in ("source", "rate oracle"))

    def _complete_mqtt_arguments(self, document: Document) -> bool:
        text_before_cursor: str = document.text_before_cursor
        return text_before_cursor.startswith("mqtt ")

    def get_completions(self, document: Document, complete_event: CompleteEvent):
        """
        Get completions for the current scope. This is the defining function for the completer
        :param document:
        :param complete_event:
        """
        if self._complete_pmm_script_files(document):
            for c in self._py_file_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_script_strategy_files(document):
            for c in self._script_strategy_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_script_strategy_config(document):
            for c in self._scripts_config_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_strategy_v2_files_with_config(document):
            for c in self._strategy_v2_create_config_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_controllers_config(document):
            for c in self._controller_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_paths(document):
            for c in self._path_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_strategies(document):
            for c in self._strategy_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_chain(document):
            for c in self._gateway_chain_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_network(document):
            for c in self._gateway_network_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_wallet_addresses(document):
            for c in self._gateway_wallet_address_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_exchange_amm_connectors(document):
            if self._complete_spot_exchanges(document):
                for c in self._spot_exchange_completer.get_completions(document, complete_event):
                    yield c
            else:
                for c in self._exchange_amm_completer.get_completions(document, complete_event):
                    yield c

        elif self._complete_exchange_clob_connectors(document):
            if self._complete_spot_exchanges(document):
                for c in self._spot_exchange_completer.get_completions(document, complete_event):
                    yield c
            elif self._complete_derivatives(document):
                for c in self._derivative_exchange_completer.get_completions(document, complete_event):
                    yield c
            else:
                for c in self._exchange_clob_completer.get_completions(document, complete_event):
                    yield c

        elif self._complete_spot_exchanges(document):
            for c in self._spot_exchange_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_trading_timeframe(document):
            for c in self._trading_timeframe_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_connect_options(document):
            for c in self._connect_option_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_export_options(document):
            for c in self._export_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_balance_limit_exchanges(document):
            for c in self._connect_option_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_balance_options(document):
            for c in self._balance_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_history_arguments(document):
            for c in self._history_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_connect_arguments(document):
            for c in self._gateway_connect_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_connector_tokens_arguments(document):
            for c in self._gateway_connector_tokens_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_balance_arguments(document):
            for c in self._gateway_balance_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_approve_tokens_arguments(document):
            for c in self._gateway_approve_tokens_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_arguments(document):
            for c in self._gateway_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_gateway_config_arguments(document):
            for c in self._gateway_config_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_derivatives(document):
            if self._complete_exchanges(document):
                for c in self._derivative_exchange_completer.get_completions(document, complete_event):
                    yield c
            elif "(Exchange/CLOB)" in self.prompt_text:
                for c in self._derivative_completer.get_completions(document, complete_event):
                    yield c
            else:
                for c in self._derivative_completer.get_completions(document, complete_event):
                    yield c

        elif self._complete_exchanges(document):
            for c in self._exchange_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_trading_pairs(document):
            for c in self._trading_pair_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_command(document):
            for c in self._command_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_configs(document):
            for c in self._config_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_options(document):
            for c in self._option_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_rate_oracle_source(document):
            for c in self._rate_oracle_completer.get_completions(document, complete_event):
                yield c

        elif self._complete_mqtt_arguments(document):
            for c in self._mqtt_completer.get_completions(document, complete_event):
                yield c

        else:
            text_before_cursor: str = document.text_before_cursor
            try:
                first_word: str = text_before_cursor[0:text_before_cursor.index(' ')]
            except ValueError:
                return
            subcommand_completer: Completer = self.get_subcommand_completer(first_word)
            if complete_event.completion_requested or self._complete_subcommand(document):
                for c in subcommand_completer.get_completions(document, complete_event):
                    yield c


def load_completer(hummingbot_application):
    return HummingbotCompleter(hummingbot_application)



from typing import List, Tuple, cast

from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.gateway.amm.gateway_ethereum_amm import GatewayEthereumAMM
from hummingbot.connector.gateway.gateway_price_shim import GatewayPriceShim
from hummingbot.strategy.cross_exchange_market_making.cross_exchange_market_making import (
    CrossExchangeMarketMakingStrategy,
    LogOption,
)
from hummingbot.strategy.maker_taker_market_pair import MakerTakerMarketPair
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple


def start(self):
    c_map = self.strategy_config_map
    maker_market = c_map.maker_market.lower()
    taker_market = c_map.taker_market.lower()
    raw_maker_trading_pair = c_map.maker_market_trading_pair
    raw_taker_trading_pair = c_map.taker_market_trading_pair
    status_report_interval = self.client_config_map.strategy_report_interval
    debug_price_shim = c_map.debug_price_shim

    try:
        maker_trading_pair: str = raw_maker_trading_pair
        taker_trading_pair: str = raw_taker_trading_pair
        maker_assets: Tuple[str, str] = self._initialize_market_assets(maker_market, [maker_trading_pair])[0]
        taker_assets: Tuple[str, str] = self._initialize_market_assets(taker_market, [taker_trading_pair])[0]
    except ValueError as e:
        self.notify(str(e))
        return

    market_names: List[Tuple[str, List[str]]] = [
        (maker_market, [maker_trading_pair]),
        (taker_market, [taker_trading_pair]),
    ]

    self._initialize_markets(market_names)
    maker_data = [self.markets[maker_market], maker_trading_pair] + list(maker_assets)
    taker_data = [self.markets[taker_market], taker_trading_pair] + list(taker_assets)
    maker_market_trading_pair_tuple = MarketTradingPairTuple(*maker_data)
    taker_market_trading_pair_tuple = MarketTradingPairTuple(*taker_data)
    self.market_trading_pair_tuples = [maker_market_trading_pair_tuple, taker_market_trading_pair_tuple]
    self.market_pair = MakerTakerMarketPair(maker=maker_market_trading_pair_tuple, taker=taker_market_trading_pair_tuple)

    if taker_market in AllConnectorSettings.get_gateway_amm_connector_names():
        if debug_price_shim:
            amm_connector: GatewayEthereumAMM = cast(GatewayEthereumAMM, self.market_pair.taker.market)
            GatewayPriceShim.get_instance().patch_prices(
                maker_market,
                maker_trading_pair,
                amm_connector.connector_name,
                amm_connector.chain,
                amm_connector.network,
                taker_trading_pair
            )

    strategy_logging_options = (
        LogOption.CREATE_ORDER,
        LogOption.ADJUST_ORDER,
        LogOption.MAKER_ORDER_FILLED,
        LogOption.REMOVING_ORDER,
        LogOption.STATUS_REPORT,
        LogOption.MAKER_ORDER_HEDGED
    )
    self.strategy = CrossExchangeMarketMakingStrategy()
    self.strategy.init_params(
        config_map=c_map,
        market_pairs=[self.market_pair],
        status_report_interval=status_report_interval,
        logging_options=strategy_logging_options,
        hb_app_notification=True,
    )



from decimal import Decimal
from typing import Any, Dict, List, Tuple

from pydantic import ConfigDict, SecretStr
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

# Bybit fees: https://help.bybit.com/hc/en-us/articles/360039261154
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.0006"),
    taker_percent_fee_decimal=Decimal("0.0001"),
)

CENTRALIZED = True

EXAMPLE_PAIR = "BTC-USD"


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    contract_type = exchange_info.get("contractType")
    status = exchange_info.get("status")
    valid = (status is not None and contract_type is not None
             and status in ["Trading", "Settling"] and contract_type in ["LinearPerpetual", "InversePerpetual"])
    return valid


def get_linear_non_linear_split(trading_pairs: List[str]) -> Tuple[List[str], List[str]]:
    linear_trading_pairs = []
    non_linear_trading_pairs = []
    for trading_pair in trading_pairs:
        if is_linear_perpetual(trading_pair):
            linear_trading_pairs.append(trading_pair)
        else:
            non_linear_trading_pairs.append(trading_pair)
    return linear_trading_pairs, non_linear_trading_pairs


def is_linear_perpetual(trading_pair: str) -> bool:
    """
    Returns True if trading_pair is in USDT(Linear) Perpetual
    """
    _, quote_asset = split_hb_trading_pair(trading_pair)
    return quote_asset == "USDT"


def get_next_funding_timestamp(current_timestamp: float) -> float:
    # On ByBit Perpetuals, funding occurs every 8 hours at 00:00UTC, 08:00UTC and 16:00UTC.
    # Reference: https://help.bybit.com/hc/en-us/articles/360039261134-Funding-fee-calculation
    int_ts = int(current_timestamp)
    eight_hours = 8 * 60 * 60
    mod = int_ts % eight_hours
    return float(int_ts - mod + eight_hours)


class BybitPerpetualConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="bybit_perpetual", client_data=None)
    bybit_perpetual_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Bybit Perpetual API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    bybit_perpetual_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Bybit Perpetual secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="bybit_perpetual")


KEYS = BybitPerpetualConfigMap.construct()

OTHER_DOMAINS = ["bybit_perpetual_testnet"]
OTHER_DOMAINS_PARAMETER = {"bybit_perpetual_testnet": "bybit_perpetual_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"bybit_perpetual_testnet": "BTC-USDT"}
OTHER_DOMAINS_DEFAULT_FEES = {
    "bybit_perpetual_testnet": TradeFeeSchema(
        maker_percent_fee_decimal=Decimal("-0.00025"),
        taker_percent_fee_decimal=Decimal("0.00075"),
    )
}


class BybitPerpetualTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="bybit_perpetual_testnet", client_data=None)
    bybit_perpetual_testnet_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Bybit Perpetual Testnet API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    bybit_perpetual_testnet_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Bybit Perpetual Testnet secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="bybit_perpetual_testnet")


OTHER_DOMAINS_KEYS = {
    "bybit_perpetual_testnet": BybitPerpetualTestnetConfigMap.construct()
}
